# Copyright (c) 2024 Matthew Baker.  All rights reserved.  Licenced under the
# Apache Licence 2.0.  See LICENSE file
import unittest
import shutil
import os
from typing import List
from argparse import Namespace
import configurematic


def read_file(filename):
    lines: List[str] = []
    with open(filename, "r") as f:
        for line in f:
            lines.append(line)
    return lines


class TypedDictConfigurematic(unittest.TestCase):

    def check(self, resursive: bool):
        self.assertTrue(os.path.isfile("out/testdata/dir1/file1"))
        self.assertTrue(os.path.isfile("out/testdata/dir1/dir2/file2"))
        self.assertTrue(os.path.isfile("out/testdata/dir1/dir2/file3"))
        self.assertFalse(os.path.isfile("out/testdata/dir1/dir2/file4"))

        self.assertFalse(os.path.isfile("out/testdata/dir1/file1.old"))
        self.assertFalse(os.path.isfile("out/testdata/dir1/dir2/file2.old"))
        self.assertFalse(os.path.isfile("out/testdata/dir1/dir2/file3.old"))

        file1 = read_file("out/testdata/dir1/file1")
        self.assertEqual(len(file1), 6)
        self.assertEqual(file1[0].strip(), 'ABC')
        self.assertEqual(file1[1].strip(), 'XXX="abc"')
        self.assertEqual(file1[2].strip(), 'YYY=true')
        self.assertEqual(file1[3].strip(), 'XXXYYY="abctrue"')
        self.assertEqual(file1[4].strip(), 'XXXsYYY="abc true"')
        if (resursive):
            self.assertEqual(file1[5].strip(), 'AAA=abc')
        else:
            self.assertEqual(file1[5].strip(), 'AAA=__VAR1__')

        file2 = read_file("out/testdata/dir1/dir2/file2")
        self.assertEqual(len(file2), 3)
        self.assertEqual(file2[0].strip(), 'ABC')
        self.assertEqual(file2[1].strip(), 'XXX="abc"')
        self.assertEqual(file2[2].strip(), 'ZZZ=123')

        file3 = read_file("out/testdata/dir1/dir2/file3")
        self.assertEqual(len(file3), 3)
        self.assertEqual(file3[0].strip(), 'ABC')
        self.assertEqual(file3[1].strip(), 'YYY=true')
        self.assertEqual(file3[2].strip(), 'ZZZ=123')

    def test_nonrecursive(self):
        if (os.path.isdir("out")):
            shutil.rmtree("out")
        args = Namespace()
        args.files = "testdata/files.txt"
        args.conf_file = "testdata/conf1.txt"
        args.outdir = "out"
        args.prefix = "example-"
        args.backup_ext = ".old"
        args.delimeter = "__"
        args.silent = True
        args.recursive = False
        files = configurematic.get_filenames(args.files)

        configurematic.configure(files, args)

        self.check(False)

        configurematic.configure(files, args)
        self.assertTrue(os.path.isfile("out/testdata/dir1/file1.old"))
        self.assertTrue(os.path.isfile("out/testdata/dir1/dir2/file2.old"))
        self.assertTrue(os.path.isfile("out/testdata/dir1/dir2/file3.old"))

        if (os.path.isdir("out")):
            shutil.rmtree("out")

    def test_recursive(self):
        if (os.path.isdir("out")):
            shutil.rmtree("out")
        args = Namespace()
        args.files = "testdata/files.txt"
        args.conf_file = "testdata/conf1.txt"
        args.outdir = "out"
        args.prefix = "example-"
        args.backup_ext = ".old"
        args.delimeter = "__"
        args.silent = True
        args.recursive = True
        files = configurematic.get_filenames(args.files)

        configurematic.configure(files, args)

        self.check(True)

        configurematic.configure(files, args)
        self.assertTrue(os.path.isfile("out/testdata/dir1/file1.old"))
        self.assertTrue(os.path.isfile("out/testdata/dir1/dir2/file2.old"))
        self.assertTrue(os.path.isfile("out/testdata/dir1/dir2/file3.old"))

        if (os.path.isdir("out")):
            shutil.rmtree("out")



# Copyright (c) 2024 Matthew Baker.  All rights reserved.  Licenced under the
# Apache Licence 2.0.  See LICENSE file
import argparse
import sys
import os
import re
import shutil
from typing import Mapping, List, Dict
from dotenv import dotenv_values

MAX_ITERATIONS = 10


def get_var_names(conffile: str) -> List[str]:
    varnames: List[str] = []
    with open(conffile, "r") as f:
        incl = re.compile(r"^[A-Za-z_][A-Za-z0-9_]+\s*=")
        for line in f:
            line = line.strip()
            if (line == "" or not incl.match(line)):
                continue
            parts = line.split("=", 2)
            varnames.append(parts[0])
    return varnames


def get_filenames(listfile: str) -> List[str]:
    """
    Splits input file into an array, one value per row.
    Leading and trailing whitespacve are ignored as are any lines
    that start in # (after stripping whitespace)
    """
    files: List[str] = []
    incl = re.compile(r"^\s*[^#]")
    with open(listfile, "r") as f:
        for line in f:
            line = line.strip()
            if (line != "" and incl.match(line)):
                files.append(line)
    return files


def find_all(a_str, sub):
    positions: List[int] = []
    start = 0
    while True:
        start = a_str.find(sub, start)
        if start == -1:
            return positions
        positions.append(start)
        start += len(sub)


def append_fragments(fragments: List[str],
                     line: str,
                     templates: Mapping[str, str],
                     start: int,
                     pos: int):
    end = pos
    fragment1 = line[start:pos]
    fragment2 = None
    for name in templates:
        if fragment1.startswith(name):
            end = len(name)
            if (end < len(fragment1)):
                fragment2 = fragment1[end:]
                fragment1 = fragment1[:end]
            break
    fragments.append(fragment1)
    if (fragment2 is not None):
        fragments.append(fragment2)


def splitOnVariables(line: str, templates: Mapping[str, str]):
    positions: List[int] = []
    for name in templates:
        positions = [*positions, *find_all(line, name)]
    positions.sort()
    start = 0
    fragments: List[str] = []
    for pos in positions:
        if pos == start:
            continue
        append_fragments(fragments, line, templates, start, pos)
        start = pos
    if (start < len(line)):
        append_fragments(fragments, line, templates, start, len(line))
    return fragments


def replace(filename: str,
            variables: Mapping[str, str | None],
            templates: Mapping[str, str],
            args: argparse.Namespace):

    basename = os.path.basename(filename)
    dirname = os.path.dirname(filename)
    if (dirname == ""):
        template_filename = args.prefix + basename
    else:
        template_filename = os.path.join(dirname, args.prefix + basename)
    output_filename = os.path.join(args.outdir, filename)
    backup_filename = output_filename + args.backup_ext
    output_exists = os.path.isfile(output_filename)

    if (output_exists and args.backup_ext != ""):
        shutil.copyfile(output_filename, backup_filename)

    os.makedirs(os.path.dirname(output_filename), exist_ok=True)

    with open(output_filename, "w") as outfile, \
            open(template_filename, "r") as infile:
        for line in infile:
            replaced = False
            first = True
            count = 0
            while (first or (args.recursive and replaced)):
                count += 1
                if (count >= MAX_ITERATIONS):
                    break  # otherwise we may have an infinite loop
                first = False
                replaced = False
                fragments = splitOnVariables(line, templates)
                new_fragments: List[str] = []
                for frag in fragments:
                    if (frag in templates):
                        replaced = True
                        name = templates[frag]
                        val = ""
                        val_or_none = variables[name]
                        if (val_or_none is not None):
                            val = val_or_none
                        frag = val
                    new_fragments.append(frag)
                newline = "".join(new_fragments)
                line = newline
            outfile.write(newline)


def configure(files: List[str], args: argparse.Namespace):
    vars = dotenv_values(args.conf_file)
    varnames = get_var_names(args.conf_file)
    templates: Dict[str, str] = {}
    for name in varnames:
        templates[args.delimeter+name+args.delimeter] = name

    for file in files:
        if (not args.silent):
            print(file, "...")
        replace(file, vars, templates, args)


def main():

    parser = argparse.ArgumentParser(
        prog='configurematic',
        description='Put configuration from one file into multiple files',
    )
    parser.add_argument('filenames', nargs='*',
                        help='Template files (without prefix)')
    parser.add_argument('-c', '--conf-file', default="config.env",
                        help='Env file with variable settings')
    parser.add_argument('-p', '--prefix', default="example-",
                        help='Template files begin with this')
    parser.add_argument('-f', '--files',
                        help='List of template files (without prefix)')
    parser.add_argument('-o', '--outdir', nargs='?', default=".",
                        help='Where to write output files')
    parser.add_argument('-b', '--backup-ext', default=".old",
                        help='For previous version of output files')
    parser.add_argument('-q', '--silent', action='store_true',
                        help="Don't print anything to stdout")
    parser.add_argument('-r', '--recursive', action='store_true',
                        help="Allow template variable to reference others")
    parser.add_argument('-d', '--delimeter', default="__",
                        help="Variable instances have this prefis and suffix")
    args = parser.parse_args()
    if (len(args.filenames) == 0 and args.files is None):
        print("Please either specify filenames or -f", file=sys.stderr)
        parser.print_usage()
        sys.exit(1)

    files = args.filenames if len(args.filenames) > 0 \
        else get_filenames(args.files)

    configure(files, args)


if __name__ == "__main__":
    main()
